#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/cons.h>
#include <asm/io.h>

#include <rtl_sched.h>
#include <rtl_fifo.h>

#include <linux/string.h>

// erster Parallelport:  0x378
// zweiter Parallelport: 0x278
#define LPT_PORT 0x278

// Enable-Signal liegt auf D7 des Parallelports
#define ENABLE	0x0080

#define STATUS_STOP		0
#define STATUS_SPEEDUP	1
#define STATUS_RUNNING	2
#define STATUS_SPEEDDN	3

// Globale Variablen
pthread_t controlthread;
pthread_t stepperthread;

int  status	 = STATUS_STOP;
long changespeed = 0;

// Funktionsprototypen
void* stepper_control(void* arg);
void* stepper_out(void* arg);


// Steuer-Thread
void* stepper_control(void* arg) 
{
	char *buf[1024];
	int  count	= 255;

	pthread_make_periodic_np (pthread_self(), gethrtime(), 10000000);

	while(1)
	{
		// Daten vom FIFO einlesen
		int l = rtf_get(0, buf, 1024);
		
		// Daten vom FIFO interpretieren
		if(l>0)
		{
			if(strncmp("start",(char *)buf,5)==0 && status == STATUS_STOP)
			{
				rtl_printf("starting\n");

				// Thread starten
				pthread_wakeup_np(stepperthread);

				// Status weitersetzen			
				status = STATUS_SPEEDUP;
				count = 255;
			}
			else if(strncmp("emerg",(char *)buf,5)==0 && status > STATUS_STOP)
			{
				rtl_printf("emergency stop\n");

				// Status weitersetzen			
				status = STATUS_STOP;
			}
			else if(strncmp("stop",(char *)buf,4)==0 && status > STATUS_STOP)
			{
				rtl_printf("stopping\n");

				// Status weitersetzen			
				status = STATUS_SPEEDDN;
			}
		}
		
		// Beschleunigungskurve berechnen (quadratische Gleichung)
		if(status == STATUS_SPEEDUP)
		{
			if(count == 0)
			{
				status = STATUS_RUNNING;
			}
			else
			{
				count=count-1;
				changespeed = (count*count*61)+1000000;
			}
		}
		else if(status == STATUS_SPEEDDN)
		{
			if(count == 255)
			{
				status = STATUS_STOP;
			}
			else
			{
				count=count+1;
				changespeed = (count*count*61)+1000000;
			}
		}
		
		// Warten bis zum nchsten Aufruf
		pthread_wait_np();
	}

	return 0;
}

// Ausgabe-Thread
void* stepper_out(void* arg) 
{
	// Daten fr die Ausgabe der Bitmuster	
	int cnt = 0;
	const int val[] = {0x01,0x05,0x04,0x06,0x02,0x0A,0x08,0x09};

	// Thread als periodisch markieren und schlafenlegen
	pthread_make_periodic_np (pthread_self(), gethrtime(), 5000000);
	pthread_suspend_np(pthread_self());

    while(1)
    {
		// Wenn status auf STOP gesetzt, schlafen legen
		if(status == STATUS_STOP)
		{
			outb(0,LPT_PORT);
			pthread_suspend_np(pthread_self());
		}
	
		// Auf nderung der Geschwindigkeit reagieren
		if(changespeed>0)
		{
			pthread_make_periodic_np (pthread_self(), gethrtime(), changespeed);
			changespeed = 0;
		}	
	
		// Bitmuster ausgeben, um den Stepper einen Schritt weiterzusetzen
		outb(val[cnt%8]|ENABLE,LPT_PORT);
		
		// Zhler auf das nchste Bitmuster
		cnt++;
		
		// Warten bis zur nchsten Ausgabe
		pthread_wait_np();
    }
    
    return 0;
}


int init_module(void)
{
	// Hello-Message ausgeben (dmesg)
	rtl_printf("Stepper Control is running!\n");

	// Port in definierten Zustand bringen
	outb(0,LPT_PORT);

	// FIFO 0 (/dev/rtf0) initialisieren
	rtf_create(0, 1024*1024);

	// Threads initialisieren
    pthread_create(&stepperthread, NULL, stepper_out, (void*) 0);
    pthread_create(&controlthread, NULL, stepper_control, (void*) 0);

    return 0;
}

void cleanup_module(void)
{
	rtl_printf("Stepper Control is shutting down!\n");

	// Threads anhalten
    pthread_delete_np(stepperthread);
    pthread_delete_np(controlthread);

	// Enable zurcksetzen
	outb(0,LPT_PORT);
	
	// Eingabe-FIFO anhalten
	rtf_destroy(0);

}
